\ decoder.fs
\ Hannes Hauer
\ Simon Himmelbauer
\ Siegfried Oleg Pammer
\ Description: instruction decoder

needs constants.fs

\ input
0 value image
0 value pc
0 value codesize
0 value nibble1
0 value nibble2
0 value nibble3
0 value nibble4

\ output
0 value imm       \ N / NN / NNN
0 value register1 \ X
0 value register2 \ Y
0 value flag

: load-image ( addr length -- )
  slurp-file to codesize
  to image
;

: combine-nibble234
  nibble2 8 lshift
  nibble3 4 lshift
  or
  nibble4
  or
;

: combine-nibble34
  nibble3 4 lshift
  nibble4
  or
;

: chp8-decode-error
  s\" CHIP-8 decode error at " type
  pc . nibble1 . nibble2 . nibble3 . nibble4 .
  -1
;

: decode-0XXX
  nibble2 0= if
    combine-nibble34
    case
      $E0 of INST_CLEAR endof \ 00E0: clear
      $EE of INST_RETURN endof \ 00EE: return
      chp8-decode-error swap
    endcase
  else
    \ 0NNN: call NNN
    \ combine-nibble234 to imm
    INST_NOP \ INST_CALL
  endif
;

: decode-1XXX \ 1NNN: goto NNN
  combine-nibble234 to imm
  INST_GOTO
;

: decode-2XXX \ 2NNN: *(0xNNN)()
  combine-nibble234 to imm
  INST_CALL
;

: decode-3XXX \ 3XNN: if (Vx == NN)
  nibble2 to register1
  combine-nibble34 to imm
  FLAG_EQUAL to flag
  INST_CMP_IMM
;

: decode-4XXX \ 4XNN: if (Vx != NN)
  nibble2 to register1
  combine-nibble34 to imm
  FLAG_NEQUAL to flag
  INST_CMP_IMM
;

: decode-5XXX \ 5XY0: if (Vx == Vy)
  nibble2 to register1
  nibble3 to register2
  FLAG_EQUAL to flag
  INST_CMP_REG
;

: decode-6XXX \ 6XNN: Vx = NN
  nibble2 to register1
  combine-nibble34 to imm
  INST_MOV_IMM
;

: decode-7XXX \ 7XNN: Vx += NN
  nibble2 to register1
  combine-nibble34 to imm
  FLAG_ADD to flag
  INST_MOV_IMM
;

: decode-8XXX \ 8XY_: Vx op= Vy
  nibble2 to register1
  nibble3 to register2
  
  INST_MOV_REG
  nibble4
  case
    $0 of FLAG_NONE    to flag endof \ Vx = Vy
    $1 of FLAG_BIT_OR  to flag endof \ Vx |= Vy
    $2 of FLAG_BIT_AND to flag endof \ Vx &= Vy
    $3 of FLAG_BIT_XOR to flag endof \ Vx ^= Vy
    $4 of FLAG_ADD     to flag endof \ Vx += Vy
    $5 of FLAG_SUB     to flag endof \ Vx -= Vy
    $6 of FLAG_RSHIFT  to flag endof \ Vx >>= 1
    $7 of FLAG_SUB_INV to flag endof \ Vx = Vy - Vx
    $e of FLAG_LSHIFT  to flag endof \ Vx <<= 1
    swap drop chp8-decode-error swap
  endcase
;

: decode-9XXX \ 9XY0: if (Vx != Yy)
  nibble2 to register1
  nibble3 to register2
  FLAG_NEQUAL to flag
  INST_CMP_REG
;

: decode-AXXX \ ANNN: I = NNN
  combine-nibble234 to imm
  INST_MOV_IMM_I
;

: decode-BXXX \ BNNN: PC = V0 + NNN
  combine-nibble234 to imm
  INST_JUMP
;

: decode-CXXX \ CXNN: Vx = rand() & NN
  nibble2 to register1
  combine-nibble34 to imm
  INST_RAND
;

: decode-DXXX \ DXYN: draw(Vx, Vy, N)
  nibble2 to register1
  nibble3 to register2
  nibble4 to imm
  INST_DRAW
;

: decode-EXXX \ EX__: if (key() op Vx)
  nibble2 to register1
  
  INST_CMP_KEY
  combine-nibble34
  case
    $9E of FLAG_EQUAL to flag endof
    $A1 of FLAG_NEQUAL to flag endof
    swap drop chp8-decode-error swap
  endcase
;

: decode-FXXX
  nibble2 to register1
  combine-nibble34
  case
    $07 of INST_GET_TIME endof
    $0A of INST_GET_KEY endof
    $15 of
      FLAG_DELAY to flag
      INST_SET_TIME
    endof
    $18 of
      FLAG_SOUND to flag
      INST_SET_TIME
    endof
    $1E of
      FLAG_ADD to flag
      INST_MOV_REG_I
    endof
    $29 of INST_SET_SPRITE endof
    $33 of
      FLAG_BCD to flag
      INST_MOV_REG_I
    endof
    $55 of INST_REG_DUMP endof
    $65 of INST_REG_LOAD endof
    chp8-decode-error swap
  endcase
;

create decoders
  ' decode-0XXX , ' decode-1XXX , ' decode-2XXX , ' decode-3XXX , 
  ' decode-4XXX , ' decode-5XXX , ' decode-6XXX , ' decode-7XXX , 
  ' decode-8XXX , ' decode-9XXX , ' decode-AXXX , ' decode-BXXX , 
  ' decode-CXXX , ' decode-DXXX , ' decode-EXXX , ' decode-FXXX ,

\ Decode CHIP-8 instruction based on first nibble
: decode-instruction
  0 to imm
  0 to register1
  0 to register2
  FLAG_NONE to flag
  nibble1 cells decoders + perform
;
